#include <stdio.h>
#include <time.h>
#include <omp.h>

extern double wtime();
#ifndef NUM_STEPS
#define NUM_STEPS 10000000000
#endif

long long num_steps = NUM_STEPS;
double step;

#if !defined(CALC_PI) \
	&& !defined(CALC_PI_NO_PARALLEL) \
	&& !defined(CALC_PI_NO_VEC) \
	&& !defined(CALC_PI_NO_PARALLEL_NO_VEC) \
	&& !defined(TEST_BYTES) 	
#define CALC_PI
#endif

//--------------------------------------------------------------------
//                 CalcSum() 
//--------------------------------------------------------------------
double CalcSum(long long  i,double sum)
{
  double x;
  x = (i + .5)*step;
  sum = sum + 4.0/(1.+ x*x);
  return sum;
}

#ifdef CALC_PI
//--------------------------------------------------------------------
//                 CalcPi() 
//--------------------------------------------------------------------
double CalcPi()
{
  double pi, sum=0.0;
  long long  i;
	
#pragma omp parallel for reduction(+:sum)
  for (i=0; i<num_steps; i++)
  {
    sum = CalcSum(i,sum);
  }

  pi = sum*step;
  return pi;
}
#endif


#ifdef CALC_PI_NO_PARALLEL
//--------------------------------------------------------------------
//                 CalcPiNoParallel() 
//--------------------------------------------------------------------
double CalcPiNoParallel()
{
  double pi, sum=0.0;
  long long i;

  for (i=0; i<num_steps; i++)
  {
    sum = CalcSum(i,sum);
  }

  pi = sum*step;
  return pi;
}
#endif

#ifdef CALC_PI_NO_VEC
//--------------------------------------------------------------------
//                 CalcPiNoVec() 
//--------------------------------------------------------------------
double CalcPiNoVec()
{
  double pi, sum=0.0;
  long long i;
	

#pragma omp parallel for reduction(+:sum)
#pragma novector
  for (i=0; i<num_steps; i++)
  {
    sum = CalcSum(i,sum);
  }

  pi = sum*step;
  return pi;
}
#endif

#ifdef CALC_PI_NO_PARALLEL_NO_VEC
//--------------------------------------------------------------------
//                 CalcPiNoParallelNoVec() 
//--------------------------------------------------------------------
double CalcPiNoParallelNoVec()
{
  double pi, sum=0.0;
  long long i;
	
#pragma novector
  for (i=0; i<num_steps; i++)
  {
    sum = CalcSum(i,sum);
  }

  pi = sum*step;
  return pi;
}
#endif

//--------------------------------------------------------------------
//                 testBytes() 
//--------------------------------------------------------------------
#ifdef TEST_BYTES

#define REAL double
#define BW_ARRAY_SIZE (1000*1000*64) 
#define BW_ITERS 1000

// number of mem ops each iteration
// 1 read  + 1 write
#define OPS_PERITER 2    

// define some arrays 
// make sure they are 64 byte aligned - for fastest cache access 
REAL fc[BW_ARRAY_SIZE] __attribute__((align(64)));
REAL fd[BW_ARRAY_SIZE] __attribute__((align(64)));
double testBytes()
{
    int i,j,k;
    // use omp to scale the test across 
    // the threads requested. Need to set environment 
    // variables OMP_NUM_THREADS and KMP_AFFINITY
    for (i=0; i<BW_ITERS; i++)
    {
    //
    // copy the arrays to/from memory (2 bw ops)
    // use openmp to scale and get aggregate bw 
    //
#pragma omp parallel for
        for(k=0; k<BW_ARRAY_SIZE; k++)  
        {
           fc[k] = fd[k];
        }
    }

    double gbytes = (double)( 1.0e-9 * OPS_PERITER * BW_ITERS * 
                       BW_ARRAY_SIZE*sizeof(REAL));    

    return gbytes;
}
#endif

//--------------------------------------------------------------------
//                 main() 
//--------------------------------------------------------------------
int main(int argc, char* argv[])
{
  double start,first_start,stop;
  double pi;
  char *KMP_AFFINITY = getenv("KMP_AFFINITY");
  char *OMP_NUM_THREADS = getenv("OMP_NUM_THREADS");
  int numthreads;
  int numcores = omp_get_num_procs();

 // need to be in a parallel region to get thread nums.
 #pragma omp parallel
 #pragma omp master
  numthreads = omp_get_num_threads();
  
  
  printf("INFO - The following environment variables are set\n   KMP_AFFINITY: %s\n   OMP_NUMTHREADS: %s\n",
	        KMP_AFFINITY?KMP_AFFINITY:"Not Set",
		OMP_NUM_THREADS?OMP_NUM_THREADS:"Not Set");
	
  printf("   Num loops %lld, Num threads: %d\n   Num Cores: %d\n\n",
	 num_steps,numthreads, numcores);

  step = 1./(double)num_steps;
	
  start = wtime();
  first_start = start;
  
//--------------------------------------------------------------------
  #ifdef CALC_PI
  printf ("Running CalcPi()\n");
  pi = CalcPi();
  stop = wtime();
  printf("The value of PI is %15.12f\n",pi);
  printf("The time to calculate PI was %10.3lf seconds\n\n",stop - start);
  #endif
  
//--------------------------------------------------------------------
  #ifdef CALC_PI_NO_VEC
  start = wtime();
  printf ("Running CalcPiNoVec()\n");
  num_steps = num_steps/2;
  step = 1./(double)num_steps;
  pi = CalcPiNoVec();
  stop = wtime();
  printf("The value of PI is %15.12f\n",pi);
  printf("Time for CalcPiNoVec %10.3lf seconds\n\n",stop - start);
  #endif
  
//--------------------------------------------------------------------
  #ifdef CALC_PI_NO_PARALLEL
  start = wtime();
  printf ("Running CalcPiNoParallel()\n");
  num_steps = num_steps/4;
  step = 1./(double)num_steps;
  pi = CalcPiNoParallel();
  stop = wtime();
  printf("The value of PI is %15.12f\n",pi);
  printf("Time for CalcPiNoParallel %10.3lf seconds\n\n",stop - start);
  #endif
 
//--------------------------------------------------------------------
 #ifdef CALC_PI_NO_PARALLEL_NO_VEC
  start = wtime();
  printf ("Running CalcPiNoParallelNoVec()\n");
  num_steps = num_steps/2;
  step = 1./(double)num_steps;
  pi = CalcPiNoParallelNoVec();
  stop = wtime();
  printf("The value of PI is %15.12f\n",pi);
  printf("Time for CalcPiNoParallelNoVec %10.3lf seconds\n\n",stop - start);
  #endif
 
//--------------------------------------------------------------------
  #ifdef TEST_BYTES
  start = wtime();
  printf ("Running testBytes()\n");
  testBytes();
  stop = wtime();
  printf("Time for testBytes %10.3lf seconds\n\n",stop - start);
  #endif
 
  
  printf("Elasped time %10.3lf seconds\n",stop - first_start);
  
  return 0;
}