/* ========================================================================== *
!*
!* SAMPLE SOURCE CODE - SUBJECT TO THE TERMS OF SAMPLE CODE LICENSE AGREEMENT,
!* http://software.intel.com/en-us/articles/
!*        intel-sample-source-code-license-agreement/
!*
!* Copyright 2014 Intel Corporation
!*
!* THIS FILE IS PROVIDED "AS IS" WITH NO WARRANTIES, EXPRESS OR IMPLIED,
!* INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTY OF MERCHANTABILITY,
!* FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT OF INTELLECTUAL
!* PROPERTY RIGHTS.
!*
!* ========================================================================== */
#include <omp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <sys/time.h>

#define MYMIC mic:0

// Timer implementation
static struct timeval tstart, tstop;
void start_timer()
{
    gettimeofday(&tstart, NULL);
}

double stop_timer()
{
    gettimeofday(&tstop, NULL);
    return (tstop.tv_sec + tstop.tv_usec / 1000000.0) -
           (tstart.tv_sec + tstart.tv_usec / 1000000.0);
}

// 1D linear convolution - the actual work
__declspec(target(mic))
void convolve(float *input, float *nextInput, float *kernel, float *output,
              size_t inputSize, size_t kernelSize)
{
#pragma omp parallel for
    for (size_t i = 0; i < inputSize; i++) {
        float sum = 0;

        for (int j = 0; j < kernelSize; j++) {
            sum += kernel[j] * input[i + j];
        }
        output[i] = sum;
    }

    // Forward the trailing not yet fully convoluted elements
    if (nextInput) {
        memcpy(&nextInput[0], &input[inputSize - (kernelSize - 1)], sizeof(float) * (kernelSize - 1));
    }
}

int main(int argc, char **argv)
{
    int numThreads = 0;
    double duration = 0.0;
    FILE *fhInput = NULL;
    FILE *fhKernel = NULL;
    FILE *fhOutput = NULL;

    if (argc != 4) {
        printf("Error: Wrong use!\n");
        printf("Usage: %s <input> <kernel> <output>\n", argv[0]);
        return 1;
    }

    fhInput = fopen(argv[1], "rb");
    if (fhInput == NULL) {
        printf("Error: Cannot read input file!\n");
        return 1;
    }
    fhKernel = fopen(argv[2], "rb");
    if (fhKernel == NULL) {
        printf("Error: Cannot read kernel file!\n");
        return 1;
    }
    fhOutput = fopen(argv[3], "wb");
    if (fhOutput == NULL) {
        printf("Error: Cannot open output file!\n");
        return 1;
    }

    // Get sizes of input and kernel
    fseek(fhInput, 0, SEEK_END);
    size_t inputSize = ftell(fhInput) / sizeof(float);
    rewind(fhInput);

    if (inputSize == 0) {
        printf("Error: Input file is empty!\n");
        fclose(fhInput);
        fclose(fhKernel);
        fclose(fhOutput);
        return 1;
    }

    printf("# elements of input file: %d\n", inputSize);

    fseek(fhKernel, 0, SEEK_END);
    size_t kernelSize = ftell(fhKernel) / sizeof(float);
    rewind(fhKernel);

    if (kernelSize == 0) {
        printf("Error: Kernel file is empty!\n");
        fclose(fhInput);
        fclose(fhKernel);
        fclose(fhOutput);
        return 1;
    }

    printf("# elements of kernel file: %d\n", kernelSize);

    if (inputSize < kernelSize) {
        printf("Error: Input is smaller than kernel!\n");
        fclose(fhInput);
        fclose(fhKernel);
        fclose(fhOutput);
        return 1;
    }

    size_t chunkSize = 1024 * 1024;

    // Some assertions to make implementaion simpler!
    assert(chunkSize > kernelSize);
    assert((inputSize % chunkSize) == 0);
    const size_t inputElements = chunkSize + kernelSize - 1;

    // Allocate size for the chunks to process
    float *input1 = (float*)malloc(inputElements * sizeof(float));
    float *input2 = (float*)malloc(inputElements * sizeof(float));
    float *output = (float*)malloc(chunkSize * sizeof(float));

    // Read in entire kernel
    float *kernel = (float*)malloc(kernelSize * sizeof(float));
    size_t curElement = 0;
    while(!feof(fhKernel)) {
        size_t readItems = 0;
        curElement += fread(&kernel[curElement], sizeof(float), kernelSize, fhKernel);
    }

    // Initialize the OpenMP runtime once and inform about available threads
    kmp_set_defaults("KMP_AFFINITY=scatter");
#pragma offload target(MYMIC)
#pragma omp parallel for
    for (size_t i = 0; i < 10; i++) {
        if (i == 0) numThreads = omp_get_num_threads();
    }
    printf("Number of OpenMP threads used: %d\n", numThreads);

    // Allocate all buffers on coprocessor once and reuse them.
    // Also transfer the kernel once (it's constant)
    //
    // Note: Array input1 is twofold as it serves the host as primary buffer for
    //       the input but for the coprocessor it is used as one of the two
    //       buffers (input1 & input2). Memory area of input1 is different for
    //       the host and coprocessor!
#pragma offload_transfer target(MYMIC) \
          nocopy(input1:length(inputElements) alloc_if(1) free_if(0)) \
          nocopy(input2:length(inputElements) alloc_if(1) free_if(0)) \
          nocopy(output:length(chunkSize)     alloc_if(1) free_if(0)) \
          in(kernel:length(kernelSize)        alloc_if(1) free_if(0))

    const int signal1 = 1;
    const int signal2 = 2;

    start_timer();
    for(int i = 0; i < kernelSize - 1; i++) input1[i] = 0; // Fade in

    // Bootstrap: kick off the initial transfer
    curElement = kernelSize - 1;
    while (curElement != inputElements) {
        curElement += fread(&input1[curElement], sizeof(float), inputElements - curElement, fhInput);
    }
#pragma offload_transfer target(MYMIC) \
          in(input1 : length(inputElements) alloc_if(0) free_if(0)) // everything into input1 (incl. fade in)

    int bufferSelector = signal1;
    for (int i = 1; i < inputSize / chunkSize; i++) {
        // Read might be interrupted, hence retry till the full chunk size is
        // read
        size_t curElement = 0;
        while (curElement != chunkSize) {
            curElement += fread(&input1[curElement], sizeof(float), chunkSize - curElement, fhInput);
        }

        if (bufferSelector == signal1) {
            // Process data of signal1 and send new data as signal2
#pragma offload_transfer target(MYMIC) signal(signal2) \
          in(input1 : length(chunkSize) into(input2[kernelSize - 1:chunkSize]) alloc_if(0) free_if(0))

#pragma offload target(MYMIC) \
          in(input1 : length(0) alloc_if(0) free_if(0)) \
          in(input2 : length(0) alloc_if(0) free_if(0)) \
          in(kernel : length(0) alloc_if(0) free_if(0)) \
          out(output : length(chunkSize) alloc_if(0) free_if(0))
            convolve(input1, input2, kernel, output, chunkSize, kernelSize);

        } else {
            // Process data of signal2 and send new data as signal1
#pragma offload_transfer target(MYMIC) signal(signal1) \
          in(input1 : length(chunkSize) into(input1[kernelSize - 1:chunkSize]) alloc_if(0) free_if(0))

#pragma offload target(MYMIC) \
          in(input1 : length(0) alloc_if(0) free_if(0)) \
          in(input2 : length(0) alloc_if(0) free_if(0)) \
          in(kernel : length(0) alloc_if(0) free_if(0)) \
          out(output : length(chunkSize) alloc_if(0) free_if(0))
            convolve(input2, input1, kernel, output, chunkSize, kernelSize);
        }

        // Write might be interrupted, hence retry till the full chunk size is
        // written
        //
        // Note: The remainder of kernelSize - 1 elements is still at the
        //       coprocessor. We get it later.
        size_t elementsWritten = 0;
        while (elementsWritten != chunkSize) {
            elementsWritten += fwrite(output, sizeof(float), chunkSize - elementsWritten, fhOutput);
        }

        if (bufferSelector == signal1) {
#pragma offload_transfer target(MYMIC) wait(signal2)
        } else {
#pragma offload_transfer target(MYMIC) wait(signal1)
        }
        bufferSelector = (bufferSelector == signal1) ? signal2 : signal1;
    }

    // Handle the outstanding (last) transfer
    if (bufferSelector == signal1) {
        // Process data of signal1
#pragma offload target(MYMIC) \
        in(input1 : length(0) alloc_if(0) free_if(0)) \
        in(kernel : length(0) alloc_if(0) free_if(0)) \
        out(output : length(chunkSize) alloc_if(0) free_if(0))
        convolve(input1, NULL, kernel, output, chunkSize, kernelSize);

    } else {
        // Process data of signal2
#pragma offload target(MYMIC) \
        in(input2 : length(0) alloc_if(0) free_if(0)) \
        in(kernel : length(0) alloc_if(0) free_if(0)) \
        out(output : length(chunkSize) alloc_if(0) free_if(0))
        convolve(input2, NULL, kernel, output, chunkSize, kernelSize);
    }

    // Write might be interrupted, hence retry till the full chunk size is
    // written
    //
    // Note: The remainder of kernelSize - 1 elements will also be stored.
    size_t elementsWritten = 0;
    while (elementsWritten != chunkSize) {
        elementsWritten += fwrite(output, sizeof(float), chunkSize - elementsWritten, fhOutput);
    }

    duration = stop_timer();

    // Clean up the buffers on the coprocessor
#pragma offload_transfer target(MYMIC) \
      nocopy(input1:length(0) alloc_if(0) free_if(1)) \
      nocopy(input2:length(0) alloc_if(0) free_if(1)) \
      nocopy(output:length(0) alloc_if(0) free_if(1)) \
      nocopy(kernel:length(0) alloc_if(0) free_if(1))

    // Print results
    printf("Elapsed time = %lf seconds\n", duration);

    // Clean up
    fclose(fhInput);
    fclose(fhKernel);
    fclose(fhOutput);
    free(input1);
    free(input2);
    free(kernel);
    free(output);
    return 0;
}
