/* ========================================================================== *
!*
!* SAMPLE SOURCE CODE - SUBJECT TO THE TERMS OF SAMPLE CODE LICENSE AGREEMENT,
!* http://software.intel.com/en-us/articles/
!*        intel-sample-source-code-license-agreement/
!*
!* Copyright 2014 Intel Corporation
!*
!* THIS FILE IS PROVIDED "AS IS" WITH NO WARRANTIES, EXPRESS OR IMPLIED,
!* INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTY OF MERCHANTABILITY,
!* FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT OF INTELLECTUAL
!* PROPERTY RIGHTS.
!*
!* ========================================================================== */
#include <omp.h>
#include <stdio.h>
#include <stdlib.h>
#include <offload.h>
#include <sys/time.h>
#include <new>

#define MYMIC mic:0

// This is currently required to disable allocation/de-allocation for the
// implicit offload model when -no-offload was used. A future compiler version
// will do this automatically. (DPD200509198)
#ifndef __INTEL_OFFLOAD
#define _Offload_shared_malloc malloc
#define _Offload_shared_free free
#endif

// Timer implementation
static struct timeval tstart, tstop;
void start_timer()
{
    gettimeofday(&tstart, NULL);
}

double stop_timer()
{
    gettimeofday(&tstop, NULL);
    return (tstop.tv_sec + tstop.tv_usec / 1000000.0) -
           (tstart.tv_sec + tstart.tv_usec / 1000000.0);
}

_Cilk_shared
int init()
{
    int numThreads = 0;
    kmp_set_defaults("KMP_AFFINITY=scatter");
#pragma omp parallel for
    for (size_t i = 0; i < 10; i++) {
        if (i == 0) numThreads = omp_get_num_threads();
    }

    return numThreads;
}

class _Cilk_shared Data {
public:
    Data(size_t iS, size_t kS) : inputSize(iS), kernelSize(kS)
    {
        input = (_Cilk_shared float *)_Offload_shared_malloc((inputSize + kernelSize - 1) * sizeof(float));
        kernel = (_Cilk_shared float *)_Offload_shared_malloc(kernelSize * sizeof(float));
    }

    ~Data()
    {
        _Offload_shared_free(input);
        _Offload_shared_free(kernel);
    }

    _Cilk_shared float *input;
    _Cilk_shared float *kernel;
    size_t inputSize;
    size_t kernelSize;
};

// 1D linear convolution - the actual work
_Cilk_shared
void convolve(_Cilk_shared Data *data)
{
#if 1
    for (int run = 0; run < 200; run++) {
#pragma omp parallel for
        for (size_t i = 0; i < data->inputSize; i++) {
            float sum = 0;
            for (int j = 0; j < data->kernelSize; j++) {
                sum += data->kernel[j] * data->input[i + j];
            }
            data->input[i] = sum;
        }

        data->kernel[0:data->kernelSize] /= 10.0;
    }
#else
    for (int run = 0; run < 200; run++) {
#pragma omp parallel for
        for (size_t i = 0; i < data->inputSize; i++) {
            float sum = 0;
            sum += __sec_reduce_add(data->kernel[0:data->kernelSize] * data->input[i:data->inputSize]);
            data->input[i] = sum;
        }

        data->kernel[0:data->kernelSize] /= 10.0;
    }
#endif
}

_Cilk_shared Data * _Cilk_shared data;

int main(int argc, char **argv)
{
    int numThreads = 0;
    double duration = 0.0;
    FILE *fhInput = NULL;
    FILE *fhKernel = NULL;
    FILE *fhOutput = NULL;

    if (argc != 4) {
        printf("Error: Wrong use!\n");
        printf("Usage: %s <input> <kernel> <output>\n", argv[0]);
        return 1;
    }

    fhInput = fopen(argv[1], "rb");
    if (fhInput == NULL) {
        printf("Error: Cannot read input file!\n");
        return 1;
    }
    fhKernel = fopen(argv[2], "rb");
    if (fhKernel == NULL) {
        printf("Error: Cannot read kernel file!\n");
        return 1;
    }
    fhOutput = fopen(argv[3], "wb");
    if (fhOutput == NULL) {
        printf("Error: Cannot open output file!\n");
        return 1;
    }

    // Get sizes of input and kernel
    fseek(fhInput, 0, SEEK_END);
    size_t inputSize = ftell(fhInput) / sizeof(float);
    rewind(fhInput);

    if (inputSize == 0) {
        printf("Error: Input file is empty!\n");
        fclose(fhInput);
        fclose(fhKernel);
        fclose(fhOutput);
        return 1;
    }

    printf("# elements of input file: %d\n", inputSize);

    fseek(fhKernel, 0, SEEK_END);
    size_t kernelSize = ftell(fhKernel) / sizeof(float);
    rewind(fhKernel);

    if (kernelSize == 0) {
        printf("Error: Kernel file is empty!\n");
        fclose(fhInput);
        fclose(fhKernel);
        fclose(fhOutput);
        return 1;
    }

    printf("# elements of kernel file: %d\n", kernelSize);

    // Create the data structure.
    // Note that since it is shared it requires dedicated allocation for the
    // implicit offload model. We simply use the placement new operator.
    data = new (_Offload_shared_malloc(sizeof(Data))) _Cilk_shared Data(inputSize, kernelSize);

    // Read in input and kernel
    for(int i = 0; i < kernelSize - 1; i++) data->input[i] = 0; // Fade in

    size_t curElement = kernelSize - 1;
    while (curElement != (inputSize + kernelSize - 1)) {
        curElement += fread(&data->input[curElement], sizeof(float), (inputSize + kernelSize - 1) - curElement, fhInput);
    }

    curElement = 0;
    while (curElement != kernelSize) {
        curElement += fread(&data->kernel[curElement], sizeof(float), kernelSize - curElement, fhKernel);
    }

    // Initialize the OpenMP runtime once and inform about available threads
    numThreads = _Cilk_offload init();
    printf("Number of OpenMP threads used: %d\n", numThreads);

    start_timer();
    _Cilk_offload_to(MYMIC) convolve(data);
    duration = stop_timer();

    fwrite(data->input, sizeof(float), inputSize, fhOutput);

    // Print results
    printf("Elapsed time = %lf seconds\n", duration);

    // Clean up
    fclose(fhInput);
    fclose(fhKernel);
    fclose(fhOutput);
    _Offload_shared_free(data);
    return 0;
}
