!* ========================================================================== *
!*
!* SAMPLE SOURCE CODE - SUBJECT TO THE TERMS OF SAMPLE CODE LICENSE AGREEMENT,
!* http://software.intel.com/en-us/articles/
!*        intel-sample-source-code-license-agreement/
!*
!* Copyright 2014 Intel Corporation
!*
!* THIS FILE IS PROVIDED "AS IS" WITH NO WARRANTIES, EXPRESS OR IMPLIED,
!* INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTY OF MERCHANTABILITY,
!* FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT OF INTELLECTUAL
!* PROPERTY RIGHTS.
!*
!* ========================================================================== *

#define MYMIC mic:0

! 1D linear convolution - the actual work
!DIR$ attributes offload : mic ::convolve
subroutine convolve(input, nextInput, kernel, output, inputSize, kernelSize)
    implicit none
    integer i, j
    integer inputSize, kernelSize
    real*4 input(inputSize + kernelSize - 1),     &
&          nextInput(inputSize + kernelSize - 1), &
&          kernel(kernelSize),                    &
&          output(inputSize)
    real*4 res

!$OMP parallel do
    do i = 1, inputSize
        res = 0.0
        do j = 1, kernelSize
            res = res + kernel(j) * input(i + j - 1)
        enddo
        output(i) = res
    enddo

    ! Forward the trailing not yet fully convoluted elements
    nextInput(0:kernelSize - 1) = input(inputSize - (kernelSize - 2):kernelSize - 1)
end subroutine

program offload_explicit
    use omp_lib
    implicit none
    integer i
    integer argc
    integer inputSize, kernelSize
    integer numThreads
    character(len=256), allocatable :: argv(:)
    integer starttime, stoptime, rate
    real*4, allocatable :: input1(:), input2(:), kernel(:), output(:)
    logical inputOK, kernelOK, outputOK
    integer, parameter :: chunkSize = 1024*1024
    integer signal1, signal2
    integer inputElements
    integer bufferSelector
!DIR$ attributes offload : mic ::convolve

    argc = command_argument_count()
    argc = argc + 1
    allocate(argv(argc))
    do i = 1, argc
        call get_command_argument(i - 1, argv(i))
    end do

    if (argc .ne. 4) then
        print *, "Error: Wrong use!"
        print *, "Usage: ", argv(.true.), " <input> <kernel> <output>"
        deallocate(argv)
        stop 1
    endif

    ! Get sizes of input and kernel and open input, kernel and output file
    inquire(file=argv(2), OPENED=inputOK, SIZE=inputSize)
    inputSize = inputSize / 4;
    if (inputOK) then
        print *, "Error: Cannot read input file!"
        deallocate(argv)
        stop 1
    endif
    if (inputSize .eq. 0) then
        print *, "Error: Input file is empty!"
        deallocate(argv)
        stop 1
    endif
    print *, "# elements of input file: ", inputSize
    open(file=argv(2), unit=8, access="stream",form="unformatted")

    inquire(file=argv(3), OPENED=kernelOK, SIZE=kernelSize)
    kernelSize = kernelSize / 4;
    if (kernelOK) then
        print *, "Error: Cannot read kernel file!"
        deallocate(argv)
        close(8)
        stop 1
    endif
    if (kernelSize .eq. 0) then
        print *, "Error: Kernel file is empty!"
        deallocate(argv)
        close(8)
        stop 1
    endif
    print *, "# elements of kernel file: ", kernelSize
    if (inputSize .lt. kernelSize) then
        print *, "Error: Input is smaller than kernel!"
        deallocate(argv)
        close(8)
        stop 1
    endif
    open(file=argv(3), unit=9, access="stream",form="unformatted")

    inquire(file=argv(4), OPENED=outputOK)
    if (outputOK) then
        print *, "Error: Cannot read kernel file!"
        deallocate(argv)
        close(8)
        close(9)
        stop 1
    endif

    ! Some assertions to make implementaion simpler!
    if (.not.(chunkSize .gt. kernelSize) .or. &
&       .not.(mod(inputSize, chunkSize) .eq. 0)) then
        print *, "Error: Assertion!"
        close(8)
        close(9)
        stop 1
    endif
    inputElements = chunkSize + kernelSize - 1

    open(file=argv(4), unit=10, access="stream",form="unformatted")
    deallocate(argv)


    ! Allocate size for the chunks to process
    allocate(input1(inputElements))
    allocate(input2(inputElements))
    allocate(output(chunkSize))

    ! Read in entire kernel
    allocate(kernel(kernelSize))
    read(9) kernel
    close(9)

    ! Initialize the OpenMP runtime once and inform about available threads
    call kmp_set_defaults("KMP_AFFINITY=scatter");
!DIR$ offload target(MYMIC)
!$OMP parallel do
    do i = 1, 10
        if (i .eq. 1) then
            numThreads = OMP_GET_NUM_THREADS()
        endif
    enddo
    print *, "Number of OpenMP threads used: ", numThreads

    ! Allocate all buffers on coprocessor once and reuse them.
    ! Also transfer the kernel once (it's constant)
    !
    ! Note: Array input1 is twofold as it serves the host as primary buffer for
    !       the input but for the coprocessor it is used as one of the two
    !       buffers (input1 & input2). Memory area of input1 is different for
    !       the host and coprocessor!
!DIR$ offload_transfer target(MYMIC)                                         &
      nocopy(input1:length(inputElements) alloc_if(.true.) free_if(.false.)) &
      nocopy(input2:length(inputElements) alloc_if(.true.) free_if(.false.)) &
      nocopy(output:length(chunkSize)     alloc_if(.true.) free_if(.false.)) &
      in(kernel:length(kernelSize)        alloc_if(.true.) free_if(.false.))

    call system_clock(starttime, rate)
    do i = 1, kernelSize - 1
        input1(i) = 0.0 ! Fade in
    enddo

    ! Bootstrap: kick off the initial transfer
    read(8) input1(kernelSize:chunkSize)
!DIR$ offload_transfer target(MYMIC) &
      in(input1 : length(inputElements) alloc_if(.false.) free_if(.false.)) ! everything into input1 (incl. fade in)

    bufferSelector = signal1

    do i = 2, inputSize / chunkSize
        read(8) input1(1:chunkSize)

        if (bufferSelector .eq. signal1) then
            ! Process data of signal1 and send new data as signal2
!DIR$ offload_transfer target(MYMIC) signal(signal2) &
      in(input1:length(chunkSize) into(input2(kernelSize:inputElements)) alloc_if(.false.) free_if(.false.))

!DIR$ offload target(MYMIC)                                     &
      in(input1 : length(0) alloc_if(.false.) free_if(.false.)) &
      in(input2 : length(0) alloc_if(.false.) free_if(.false.)) &
      in(kernel : length(0) alloc_if(.false.) free_if(.false.)) &
      out(output : length(chunkSize) alloc_if(.false.) free_if(.false.))
            call convolve(input1, input2, kernel, output, chunkSize, kernelSize)
        else
            ! Process data of signal2 and send new data as signal1
!DIR$ offload_transfer target(MYMIC) signal(signal1) &
      in(input1:length(chunkSize) into(input1(kernelSize:inputElements)) alloc_if(.false.) free_if(.false.))

!DIR$ offload target(MYMIC)                                       &
      in(input1 : length(0) alloc_if(.false.) free_if(.false.)) &
      in(input2 : length(0) alloc_if(.false.) free_if(.false.)) &
      in(kernel : length(0) alloc_if(.false.) free_if(.false.)) &
      out(output : length(chunkSize) alloc_if(.false.) free_if(.false.))
            call convolve(input2, input1, kernel, output, chunkSize, kernelSize)
        endif

        ! Write data to file
        !
        ! Note: The remainder of kernelSize - 1 elements is still at the
        !       coprocessor. We get it later.
        write(10) output

        if (bufferSelector .eq. signal1) then
!DIR$ offload_transfer target(MYMIC) wait(signal2)
        else
!DIR$ offload_transfer target(MYMIC) wait(signal1)
        endif
    enddo

    ! Handle the outstanding (last) transfer
    if (bufferSelector .eq. signal1) then
        !Process data of signal1
!DIR$ offload target(MYMIC)                                     &
      in(input1 : length(0) alloc_if(.false.) free_if(.false.)) &
      in(input2 : length(0) alloc_if(.false.) free_if(.false.)) &
      in(kernel : length(0) alloc_if(.false.) free_if(.false.)) &
      out(output : length(chunkSize) alloc_if(.false.) free_if(.false.))
        call convolve(input1, input2, kernel, output, chunkSize, kernelSize)

    else
        !Process data of signal2
!DIR$ offload target(MYMIC)                                     &
      in(input1 : length(0) alloc_if(.false.) free_if(.false.)) &
      in(input2 : length(0) alloc_if(.false.) free_if(.false.)) &
      in(kernel : length(0) alloc_if(.false.) free_if(.false.)) &
      out(output : length(chunkSize) alloc_if(.false.) free_if(.false.))
        call convolve(input2, input1, kernel, output, chunkSize, kernelSize)
    endif

    call system_clock(stoptime, rate)

    ! Clean up the buffers on the coprocessor
!DIR$ offload_transfer target(MYMIC)                             &
      nocopy(input1:length(0) alloc_if(.false.) free_if(.true.)) &
      nocopy(input2:length(0) alloc_if(.false.) free_if(.true.)) &
      nocopy(output:length(0) alloc_if(.false.) free_if(.true.)) &
      nocopy(kernel:length(0) alloc_if(.false.) free_if(.true.))

    print *, "Elapsed time = ", (stoptime - starttime) / real(rate), " seconds"

    close(8)
    close(10)
    deallocate(input1)
    deallocate(input2)
    deallocate(kernel)
    deallocate(output)
end program
