/* ========================================================================== *
!*
!* SAMPLE SOURCE CODE - SUBJECT TO THE TERMS OF SAMPLE CODE LICENSE AGREEMENT,
!* http://software.intel.com/en-us/articles/
!*        intel-sample-source-code-license-agreement/
!*
!* Copyright 2014 Intel Corporation
!*
!* THIS FILE IS PROVIDED "AS IS" WITH NO WARRANTIES, EXPRESS OR IMPLIED,
!* INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTY OF MERCHANTABILITY,
!* FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT OF INTELLECTUAL
!* PROPERTY RIGHTS.
!*
!* ========================================================================== */
#include <omp.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#ifdef __MIC__
#pragma message "You are compiling for the coprocessor now!"
#endif

// Timer implementation
static struct timeval tstart, tstop;
void start_timer()
{
    gettimeofday(&tstart, NULL);
}

double stop_timer()
{
    gettimeofday(&tstop, NULL);
    return (tstop.tv_sec + tstop.tv_usec / 1000000.0) -
           (tstart.tv_sec + tstart.tv_usec / 1000000.0);
}

// 1D linear convolution - the actual work
void convolve(float *input, float *kernel, float *output,
              size_t inputSize, size_t kernelSize)
{
#pragma omp parallel for
    for (size_t i = 0; i < inputSize; i++) {
        float sum = 0;
        for (int j = 0; j < kernelSize; j++) {
            sum += kernel[j] * input[i + j];
        }
        output[i] = sum;
    }
}

int main(int argc, char **argv)
{
    int numThreads = 0;
    double duration = 0.0;
    FILE *fhInput = NULL;
    FILE *fhKernel = NULL;
    FILE *fhOutput = NULL;

    if (argc != 4) {
        printf("Error: Wrong use!\n");
        printf("Usage: %s <input> <kernel> <output>\n", argv[0]);
        return 1;
    }

    fhInput = fopen(argv[1], "rb");
    if (fhInput == NULL) {
        printf("Error: Cannot read input file!\n");
        return 1;
    }
    fhKernel = fopen(argv[2], "rb");
    if (fhKernel == NULL) {
        printf("Error: Cannot read kernel file!\n");
        return 1;
    }
    fhOutput = fopen(argv[3], "wb");
    if (fhOutput == NULL) {
        printf("Error: Cannot open output file!\n");
        return 1;
    }

    // Get sizes of input and kernel
    fseek(fhInput, 0, SEEK_END);
    size_t inputSize = ftell(fhInput) / sizeof(float);
    rewind(fhInput);

    if (inputSize == 0) {
        printf("Error: Input file is empty!\n");
        fclose(fhInput);
        fclose(fhKernel);
        fclose(fhOutput);
        return 1;
    }

    printf("# elements of input file: %d\n", inputSize);

    fseek(fhKernel, 0, SEEK_END);
    size_t kernelSize = ftell(fhKernel) / sizeof(float);
    rewind(fhKernel);

    if (kernelSize == 0) {
        printf("Error: Kernel file is empty!\n");
        fclose(fhInput);
        fclose(fhKernel);
        fclose(fhOutput);
        return 1;
    }

    printf("# elements of kernel file: %d\n", kernelSize);

    // Read in input and kernel
    float *input = (float*)malloc((inputSize + kernelSize - 1) * sizeof(float));
    float *output = (float*)malloc(inputSize * sizeof(float));

    for(int i = 0; i < kernelSize - 1; i++) input[i] = 0.0; // Fade in
    size_t curElement = kernelSize - 1;
    while(!feof(fhInput)) {
        curElement += fread(&input[curElement], sizeof(float), inputSize, fhInput);
    }

    float *kernel = (float*)malloc(kernelSize * sizeof(float));
    curElement = 0;
    while(!feof(fhKernel)) {
        size_t readItems = 0;
        curElement += fread(&kernel[curElement], sizeof(float), kernelSize, fhKernel);
    }

    // Initialize the OpenMP runtime once and inform about available threads
    kmp_set_defaults("KMP_AFFINITY=scatter");
#pragma omp parallel for
    for (size_t i = 0; i < 10; i++) {
        if (i == 0) numThreads = omp_get_num_threads();
    }
    printf("Number of OpenMP threads used: %d\n", numThreads);

    start_timer();
#pragma noinline // Make sure compiler does not optmize over call
    convolve(input, kernel, output, inputSize, kernelSize);
    duration = stop_timer();

    fwrite(output, sizeof(float), inputSize, fhOutput);

    // Print results
    printf("Elapsed time = %lf seconds\n", duration);

    // Clean up
    fclose(fhInput);
    fclose(fhKernel);
    fclose(fhOutput);
    free(input);
    free(kernel);
    free(output);
    return 0;
}
